"""Test the zetup config management on zetup package's own config.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import sys
from types import ModuleType
from inspect import ismodule

from pkg_resources import (
  parse_requirements, get_distribution, DistributionNotFound)

from path import Path

import zetup
from zetup import Zetup

import pytest


def test_name(zfg):
    assert zfg.NAME == 'zetup'


def test_zetup_config(zfg, zfg_path, pkg_path, in_repo, in_site_packages):
    if in_site_packages:
        assert zfg.__name__ == 'zetup.zetup_config'
        assert zfg_path == pkg_path # / 'zetup_config'
    else:
        with pytest.raises(ImportError):
            from zetup import zetup_config

        assert isinstance(zfg, Zetup)
        assert zfg_path == pkg_path.dirname()


def test_description(zfg):
    assert zetup.__description__ == zfg.DESCRIPTION


def test_distribution(zfg, pkg_path, in_repo, in_site_packages):
    try:
        dist = get_distribution('zetup')
    except DistributionNotFound:
        if in_repo:
            dist = None
        else:
            raise
    else:
        if in_repo and not Path(dist.location).samefile(pkg_path.dirname()):
            dist = None
    assert zetup.__distribution__ == zfg.DISTRIBUTION.find(pkg_path) \
      == dist


def test_version(zfg):
    assert zetup.__version__ == zfg.VERSION


def test_requires(zfg, zfg_path, in_repo, in_site_packages):
    assert zetup.__requires__ == zfg.REQUIRES
    if in_repo:
        assert list(parse_requirements(str(zetup.__requires__))) == list(
          parse_requirements((zfg_path / 'requirements.txt').text()))


def test_extras(zfg, zfg_path, in_repo, in_site_packages):
    for extra in zetup.__extras__:
        assert zetup.__extras__[extra] == zfg.EXTRAS[extra]
        if in_repo:
            assert list(parse_requirements(str(zetup.__extras__[extra]))) \
              == list(parse_requirements(
                   (zfg_path / ('requirements.%s.txt' % extra))
                   .text()))


def test_python(zfg):
    assert ('%s.%s' % sys.version_info[:2]) in zfg.PYTHON


def test_classifiers(zfg):
    assert all('Programming Language :: Python :: %s' % version
               for version in zfg.PYTHON)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

import os

import pkg_resources
from pkg_resources import (
  get_distribution, DistributionNotFound, VersionConflict)

from .version import Version

__all__ = ['Distribution']


class Distribution(str):
    """Simple proxy to get a pkg_resources.Distribution instance
       matching dist name and version from a zetup config object.
    """
    def __new__(cls, zfg_or_name, mainpkg=None, version=None):
        if isinstance(zfg_or_name, str):
            name = zfg_or_name
        else:
            name = zfg_or_name.NAME
        return str.__new__(cls, name)

    def __init__(self, zfg_or_name, mainpkg=None, version=None):
        """Initialize with zetup config in `zfg_or_name`.

        - Supports dist name as first argument
          and remaining arguments for backwards compatibility.
        """
        if not isinstance(zfg_or_name, str):
            self.zfg = zfg_or_name
        else:
            self.zfg = None
            self.__dict__['version'] = version and Version(version)

    @property
    def version(self):
        """Get version from zetup config as :class:`zetup.Version` instance.
        """
        # first look for explicit version from __init__
        # for backwards compatibility
        return self.__dict__.get('version') \
            or self.zfg.VERSION and Version(self.zfg.VERSION)

    def find(self, modpath, raise_=True):
        """Try to find the distribution and check version.

        - Also checks if distribution is in the same directory
          as given `modpath`.
        - Automatically reinstalls zetup in develop mode
          on distribution version mismatch
          if handling zetup's own config from zetup's project src.

        :param raise_:
           Raise a ``VersionConflict``
           if version doesn't match the given one?
           If ``False`` just return ``None``.
        """
        # If no version is given (for whatever reason), just do nothing:
        if not self.version:
            return None
        try:
            dist = get_distribution(self)
        except DistributionNotFound:
            return None
        # check if distribution path matches package path
        if os.path.normcase(os.path.realpath(dist.location)) \
          != os.path.normcase(os.path.dirname(os.path.realpath(modpath))):
            return None
        if dist.parsed_version != self.version.parsed:
            message = (
                "Version of distribution %s "
                "doesn't match version %s from %s. "
                % (repr(dist), self.version, repr(self.zfg)))
            # are we handling zetup's own config?
            if self == 'zetup':
                from zetup.zetup import Zetup
                # and was it loaded from zetup's project src?
                if isinstance(self.zfg, Zetup):
                    print('zetup:', message)
                    # then automatically reinstall zetup in develop mode
                    print('zetup: Reinstalling in develop mode...')
                    import zetup.commands.dev
                    self.zfg.dev()
                    # reset setuptools distribution cache
                    pkg_resources.working_set = pkg_resources.WorkingSet(
                        pkg_resources.working_set.entries)
                    # and return updated dist
                    return get_distribution(self)

            elif raise_:
                raise VersionConflict(
                    message + "Please reinstall %s." % str.__repr__(self))
            return None
        return dist

    @property
    def py(self):
        return '%s(zfg)' % (type(self).__name__)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

import sys

import pkg_resources
#HACK: happens on setup of namespace packages
if pkg_resources.require is None:
    # might be needed by package requirements
    pkg_resources.require = pkg_resources.WorkingSet().require


from .zetup import Zetup, find_zetup_config
from .error import ZetupError
from .config import ZetupConfigNotFound
from .requires import DistributionNotFound, VersionConflict
from .resolve import resolve
from .process import Popen, call
from .object import object, meta
from .annotate import annotate
from .modules import package, toplevel, extra_toplevel
from .classpackage import classpackage
# import notebook subpackage for defining extra_toplevel below
from . import notebook


zetup = toplevel(__name__, [
    'Zetup', 'find_zetup_config',
    'ZetupError', 'ZetupConfigNotFound',
    'resolve', 'DistributionNotFound', 'VersionConflict',
    'Popen', 'call',
    'object', 'meta',
    'annotate', 'package', 'toplevel', 'extra_toplevel',
], check_requirements=False)

# can't be defined in .notebook subpackage
# because .notebook is also needed to load zetup's own config
# which is required for making extra_toplevel instantiation work
extra_toplevel(zetup, notebook.__name__, [
    'Notebook',
], check_requirements=False)
# since actual .notebook subpackage was imported before
# a manually setting of the subpackage attribute
# on the zetup toplevel wrapper is needed
sys.modules[__name__].notebook = sys.modules[notebook.__name__]


def setup_entry_point(dist, keyword='use_zetup', value=True):
    """
    Zetup's ``entry_point`` handler for the ``setup()`` process in a project's
    **setup.py**, setting all setup keyword parameters based on zetup config

    Activated with ``setup(setup_requires=['zetup'], use_zetup=True)``
    """
    assert keyword == 'use_zetup'
    if not value:
        return

    from .zetup import Zetup
    # read config from current working directory (where setup.py is run)
    zetup = Zetup()
    keywords = zetup.setup_keywords()
    for name, value in keywords.items():
        # generally, setting stuff on dist.metadata is enough (and necessary)
        setattr(dist.metadata, name, value)
        # but *pip* only works correct if some stuff is also set directly on
        # dist object (pip seems to read at least package infos somehow from
        # there)
        if name.startswith('package') or name.endswith('requires'):
            setattr(dist, name, value)

    if zetup.in_repo:
        # resolve requirements for zetup make
        resolve(['zetup[commands]>={}'.format(
            __import__('zetup').__version__)])
        import zetup.commands as _

        # make necessary files and store make result in distribution object,
        # so that files can be removed by del dist.zetup_made after setup()
        make_targets = ['VERSION', 'setup.py', 'zetup_config']
        dist.zetup_made = zetup.make(targets=make_targets)

    # finally run any custom setup hooks defined in project's zetup config
    if zetup.SETUP_HOOKS:
        sys.path.insert(0, '.')
        for hook in zetup.SETUP_HOOKS:
            modname, funcname = hook.split(':')
            mod = __import__(modname)
            for subname in modname.split('.')[1:]:
                mod = getattr(mod, subname)
            func = getattr(mod, funcname)
            func(dist)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['FILTERS', 'LOADER']

import re
from textwrap import dedent

from jinja2 import BaseLoader, TemplateNotFound


def bitbucket_rst_links(rst):
    return re.sub(
      r'<#[0-9.-]+([^>]+)>',
      lambda match: '<#rst-header-%s>' % match.group(1).lower(),
      rst)


def github_markdown_links(markdown):
    return re.sub(
      r'\[([^\]]+)\]: #(.+)',
      lambda match: '[%s]: #%s' % (
        match.group(1), match.group(2).replace('.', '').lower()),
        markdown)


FILTERS = {
  'bitbucket_rst_links': bitbucket_rst_links,
  'github_markdown_links': github_markdown_links,
  }


class ExtraTemplateLoader(BaseLoader):
    templates = {
      'bitbucket_rst': dedent("""
        {%- extends 'rst.tpl' -%}

        {% block input %}
        {% if cell.source.strip() -%}
        {% if not cell.source.startswith(('%', '!')) -%}
        .. code:: python
        {%- else -%}
        {% if cell.source.startswith('%%file') -%}
        .. code:: {{ cell.source.split('\\n')[0].rsplit('.', 1)[1] }}
        {%- else -%}
        .. code::
        {%- endif %}
        {%- endif %}

        {% if cell.outputs and not cell.source.startswith(('%', '!')) -%}
        {{ cell.source | add_prompts(cont='>>> ') | indent }}
        {%- else -%}
        {{ cell.source | indent }}
        {%- endif %}
        {%- endif %}
        {% endblock input %}

        {% block stream -%}
        .. code::

        {{ output.text | indent }}
        {%- endblock stream %}

        {% block data_text scoped -%}
        {{ output.data['text/plain'] | indent }}
        {%- endblock data_text %}

        {% block markdowncell scoped %}
        {{ super() | bitbucket_rst_links }}
        {% endblock markdowncell %}
        """),

      'github_markdown': dedent("""
        {%- extends 'markdown.tpl' -%}

        {% block input %}
        {% if cell.source.strip() %}
        {% if not cell.source.startswith(('%', '!')) -%}
        ```python
        {%- else -%}
        {% if cell.source.startswith('%%file') -%}
        ```{{ cell.source.split('\\n')[0].rsplit('.', 1)[1] }}
        {%- else -%}
        ```
        {%- endif %}
        {%- endif %}
        {% if cell.outputs and not cell.source.startswith(('%', '!')) -%}
        {{ cell.source | add_prompts(cont='>>> ') }}
        {%- else -%}
        {{ cell.source }}
        {% if not cell.outputs -%}
        ```
        {%- endif %}
        {%- endif %}
        {%- endif %}
        {% endblock input %}

        {% block stream -%}
        ```

        ```
        {{ output.text }}
        ```
        {%- endblock stream %}

        {% block data_text scoped -%}
        {{ output.data['text/plain'] }}
        ```
        {%- endblock data_text %}

        {% block markdowncell scoped %}
        {{ super() | github_markdown_links }}
        {% endblock markdowncell %}
        """),
      }

    def get_source(self, env, template):
        if template not in self.templates:
            raise TemplateNotFound(template)
        # must return (template text, template name, up_to_date() function)
        return self.templates[template], template, lambda: True


LOADER = ExtraTemplateLoader()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.commands.dev

Defines ``zetup dev`` command.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import pip

from zetup.zetup import Zetup

from zetup.commands.del_ import del_

__all__ = ['dev']


@Zetup.command(depends=['setup.py'])
def dev(zfg, args=None):
    """Install project in develop mode.
    """
    # first remove any current project installation
    del_(zfg)
    # then (re)install project in develop mode (and return pip status code)
    source = str(zfg.ZETUP_DIR)
    if zfg.EXTRAS:
        source += '[all]'
    return pip.main(['install', '--editable', source])



"""Test :mod:`zetup.object`,
   containing basic ``object``-derived class
   and ``type``-derived ``meta`` class.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from itertools import chain

import zetup


class TestMeta(object):
    """Test basic ``type``-derived :class:``zetup.meta``.
    """
    def test_class(self):
        assert issubclass(zetup.meta, type)

    def test_dir(self):
        # check that dir(meta) returns the same as dir(type)
        # (with added '__dir__' in PY2)
        assert set(dir(zetup.meta)) == set(dir(type)).union([
            '__dir__',
            'member', 'metamember',
        ])
        # and that there are no duplicate names
        assert sorted(dir(zetup.meta)) == sorted(set(dir(zetup.meta)))


class TestObject(object):
    """Test basic ``object``-derived :class:``zetup.object``.
    """
    def test_class(self):
        assert type(zetup.object) is zetup.meta

    def test_class_dir(self):
        # check that dir(zetup.object) exactly returns all member names
        # from class and metaclass level,
        # which in this case are all members from builtin object and type
        # and the added '__dir__' in PY2 (and auto-added '__weakref__')
        assert set(dir(zetup.object)) == set(chain(dir(object), dir(type), [
            '__dir__', '__weakref__',
            'member', 'metamember',
        ]))
        # and that there are no duplicate names
        assert sorted(dir(zetup.object)) == sorted(set(dir(zetup.object)))

    def test_dir(self):
        # check that dir(zetup.object()) only returns all member names
        # from class    level and auto-generated instance member names
        # (but no extra member names from metaclass level),
        # which in this case are all members from a builtin object instance,
        # the added zetup.object.__dir__ in PY2,
        # and auto-added '__weakref__', '__module__' and '__dict__'
        # from zetup.object class level
        assert set(dir(zetup.object())) == set(chain(dir(object()), [
            '__dir__', '__weakref__', '__module__', '__dict__']))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
import os

import zetup
from zetup.modules import extra_toplevel

extra_toplevel['commands'](zetup, __name__, [
    'ZetupCommandError', 'ZetupMakeError',
    'init', 'make', 'run', 'dev', 'pytest', 'tox', 'conda',
])

from path import Path

from .command import COMMANDS, command
from .error import ZetupCommandError
from .make import ZetupMakeError, make
from .run import run
from .dev import dev
from .pytest import pytest
from .tox import tox
from .conda import conda


@command
def init(name, path=None):
    """Initialize a new Zetup project in current directory or `path`.
    """
    path = Path(path if path is not None else os.getcwd())
    with open(path / 'zetuprc', 'w') as f:
        f.write("[%s]\n\n%s\n" % (name, "\n".join("%s =" % key for key in [
          'description',
          'author',
          'url',
          'license',
          'python',
          'classifiers',
          'keywords',
          ])))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

__all__ = ['Zetup', 'find_zetup_config']

import sys
import os
from importlib import import_module
from subprocess import call
try:
    from setuptools import setup, Command
except ImportError: # fallback
    # (setuptools should at least be available after package installation)
    from distutils.core import setup, Command

from .config import load_zetup_config, ZetupConfigNotFound


class Zetup(object):
    def __init__(self, ZETUP_DIR='.'):
        """Load and store zetup config from `ZETUP_DIR`
           as attributes in `self`.
        """
        load_zetup_config(ZETUP_DIR, zfg=self)

    @property
    def config(self):
        """Get the zetup config as dictionary.

        - Actually just return self.__dict__ because all attr assignments
          come from :func:`.config.load_zetup_config` in :meth:`.__init__`
          and are therefore just the config.
        """
        return self.__dict__

    def __getitem__(self, name):
        return self.config[name]

    @property
    def config_py(self):
        """Get the zetup config as Python code for writing to a .py module.
        """
        def items():
            # always start with project name
            yield "NAME = %s" % (repr(self.config['NAME']))
            for name, value in self.config.items():
                if name in [
                        'NAME', 'NOTEBOOKS',
                ] or name.startswith('ZETUP') or name.endswith('FILE'):
                    continue
                try:
                    py = value.py
                except AttributeError:
                    py = repr(value)
                yield "%s = %s" % (name, py)

        return '\n\n'.join(items())

    def setup_keywords(self):
        """Get a dictionary of `setup()` keywords generated from zetup config.
        """
        keywords = {
            'name': self.NAME,
            'description': self.DESCRIPTION,
            'author': self.AUTHOR,
            'author_email': self.EMAIL,
            'url': self.URL,
            'license': self.LICENSE,
            'setup_requires': str(self.SETUP_REQUIRES or ''),
            'install_requires': str(self.REQUIRES or ''),
            'extras_require': {
                name: str(reqs) for name, reqs in self.EXTRAS.items()},
            # always provide empty default collections
            # to make sure that any custom setup hooks
            # can always directly add stuff
            # even if no data defined in zetup config
            'packages': [],
            'package_dir': {},
            'package_data': {},
            'py_modules': self.MODULES or [],
            'entry_points': {},
            'classifiers': self.CLASSIFIERS or [],
            'keywords': self.KEYWORDS or [],
        }
        if self.EXTRAS:
            keywords['extras_require']['all'] = str(self.EXTRAS['all'])
        if self.VERSION:
            keywords['version'] = str(self.VERSION)
        if self.PACKAGES:
            namespaces = set()
            for pkg in self.PACKAGES:
                main = pkg.split('.')[0]
                if not os.path.exists(os.path.join(main, '__init__.py')):
                    namespaces.add(main)
            if namespaces:
                keywords['namespace_packages'] = list(namespaces)
            keywords['packages'] = list(map(str, self.PACKAGES))
            for pkg in self.PACKAGES:
                if pkg._path:
                    keywords['package_dir'][str(pkg)] = str(pkg._path)
        if self.ZETUP_CONFIG_PACKAGE:
            keywords['package_dir'][self.ZETUP_CONFIG_PACKAGE] = '.'
            keywords['package_data'][self.ZETUP_CONFIG_PACKAGE] \
              = self.ZETUP_DATA
        if self.SCRIPTS:
            entry_points = keywords['entry_points'].setdefault(
                'console_scripts', [])
            for name, source in self.SCRIPTS.items():
                entry_points.append('%s = %s' % (name, source))
        if self.SETUP_KEYWORDS:
            entry_points = keywords['entry_points'].setdefault(
                'distutils.setup_keywords', [])
            for name, source in self.SETUP_KEYWORDS.items():
                entry_points.append('%s = %s' % (name, source))
        cmdclasses = {}
        for cmdname in self.COMMANDS:
            cmdmethod = getattr(self, cmdname)

            class ZetupCommand(Command):
                # Must override options handling stuff from Command base...
                user_options = []

                def initialize_options(self):
                    pass

                def finalize_options(self):
                    pass

                run = staticmethod(cmdmethod)

            ZetupCommand.__name__ = cmdname
            cmdclasses[cmdname] = ZetupCommand

        keywords['cmdclass'] = cmdclasses
        return keywords

    @property
    def setup(self):
        return Setup(zfg=self)

    def __call__(self, subprocess=False, **setup_keywords):
        """Run `setup()` with generated keywords from zetup config
           and custom override `setup_keywords`.

        - If `subprocess` is True, run the dynamically generated setup.py
          (if zetup's extra 'commands' requirements are installed)
          with parameters from sys.argv in an external python process
          instead of directly calling setuptools.setup()
        """
        return self.setup(subprocess=subprocess, **setup_keywords)

    COMMANDS = []

    @classmethod
    def command(cls, name=None, args=None, depends=None):
        return CommandDeco(cls, name, args, depends)

    def __repr__(self):
        """Show class name, dist name, and project path.
        """
        return "<%s: %s (%s)>" % (
            type(self).__name__, self.NAME, self.ZETUP_DIR)


class Setup(dict):
    def __init__(self, zfg):
        dict.__init__(self, zfg.setup_keywords())
        self.zfg = zfg

    def __call__(self, subprocess=False, **keywords):
        """Run ``setup()`` with generated keywords from zetup config
           and custom override `keywords`.

        - If `subprocess` is True, run ``python setup.py`` with ``sys.argv``
          (see :meth:`Zetup.__call__` for details)
        """
        keywords = dict(self, **keywords)
        if 'make' in Zetup.COMMANDS:
            make_targets = ['VERSION', 'setup.py', 'zetup_config']
            with self.zfg.make(targets=make_targets):
                if subprocess:
                    return call([sys.executable, 'setup.py']
                                + sys.argv[1:])

                dist = setup(**keywords)
        else:
            dist = setup(**keywords)
        # to be always able to del dist.zetup_made in setup.py
        dist.zetup_made = None
        return dist


class CommandDeco(object):
    def __init__(self, zetupcls, name=None, args=None, depends=None):
        self.zetupcls = zetupcls
        self.name = name
        self.args = args
        self.make_targets = depends and list(depends)

    def __call__(self, func):
        """Add a command function as method to :class:`zetup.Zetup`
           and store its name in `cls.COMMANDS`.

        - Used in ``zetup.commands.*`` to sparate command implementations.
        """
        targets = self.make_targets

        def cmdmethod(zfg, args=None, **kwargs):
            if not targets:
                return func(zfg, args, **kwargs)
            from zetup.commands.make import make
            with make(zfg, targets=targets):
                return func(zfg, args, **kwargs)

        name = cmdmethod.__name__ = self.name or func.__name__
        cmdmethod.args = self.args

        setattr(self.zetupcls, name, cmdmethod)
        self.zetupcls.COMMANDS.append(name)
        return cmdmethod


def find_zetup_config(pkgname):
    zfg_modname = pkgname + '.zetup_config'
    try: # Already imported?
        return sys.modules[zfg_modname]
    except KeyError:
        pass
    try:
        return import_module(zfg_modname)
    except ImportError:
        pass
    # ==> no zetup config module
    # ==> assume package imported from source (repo)
    # ==> load setup config from package's parent path:
    mod = sys.modules[pkgname]
    path = os.path.dirname(os.path.dirname(os.path.realpath(mod.__file__)))
    for _ in range(pkgname.count('.')):
        # move further up the dir tree for every prefix (namespace) package
        path = os.path.dirname(path)
    try:
        return Zetup(path)
    except ZetupConfigNotFound as e:
        raise ZetupConfigNotFound(
            "No '%s.zetup_config' module and: %s" % (pkgname, e))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.object

Basic ``object``-derived class and ``type``-derived ``meta`` class
with some added features.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from itertools import chain

__all__ = ['object', 'meta']


class meta(type):
    """Basic metaclass derived from builtin ``type``,
       which adds a unified basic ``__dir__`` method for Python 2 and 3,
       which always returns all member names from metaclass and class level.
    """
    if hasattr(type, '__dir__'):  # PY3
        def __dir__(cls):
            """Get all member names from class and metaclass level.
            """
            return list(set(chain(type.__dir__(cls), dir(type(cls)))))

    else:  # PY2
        def __dir__(cls):
            """Get all member names from class and metaclass level.
            """
            return list(set(chain(
                dir(type(cls)), *(c.__dict__ for c in cls.mro()))))

    @classmethod
    def metamember(mcs, obj):
        """Decorator for adding `obj` as a member
        to the metaclass of this class ::

           class Meta(zetup.meta):
               ...

           class Class(zetup.object, metaclass=Meta):
               ...

           @Class.metamember
           def method(cls, ...):
               ...

        >>> Meta.method
        <function Meta.method>
        >>> Class.method
        <bound method Meta.method of Class>
        """
        if isinstance(obj, property):
            name = obj.fget.__name__
        else:
            try:
                name = obj.__name__
            except AttributeError:
                name = obj.__func__.__name__
            else:
                obj.__qualname__ = '%s.%s' % (
                    getattr(mcs, '__qualname__', mcs.__name__),
                    obj.__name__)
        setattr(mcs, name, obj)
        return obj

    def member(cls, obj):
        """Decorator for adding `obj` as a member to this class ::

           class Class(zetup.object):
               ...

           @Class.member
           def method(self, ...):
               ...

        >>> Class.method
        <function Class.method>
        """
        if isinstance(obj, property):
            name = obj.fget.__name__
        else:
            try:
                name = obj.__name__
            except AttributeError:
                name = obj.__func__.__name__
            else:
                obj.__qualname__ = '%s.%s' % (
                    getattr(cls, '__qualname__', cls.__name__),
                    obj.__name__)
        setattr(cls, name, obj)
        return obj


# PY2/3 compatible way to create class `object` with metaclass `meta`
clsattrs = {
    '__doc__':
    """Basic class derived from builtin ``object``,
       which adds a basic ``__dir__`` method for Python 2.
    """}
if not hasattr(object, '__dir__'):
    def __dir__(self):
        """Get all member names from instance and class level.
        """
        return list(set(chain(
            self.__dict__, *(c.__dict__ for c in type(self).mro()))))

    clsattrs['__dir__'] = __dir__

object = meta('object', (object, ), clsattrs)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

import sys
import os

from .object import object
from .zetup import find_zetup_config
from .error import ZetupError
from .package import Packages

__all__ = ['annotate', 'annotate_extra']


def annotate(pkgname, check_requirements=True, check_packages=True):
    """Find zetup config for given `pkgname`
       and add __version__, __requires__, __dist__, __description__,
       __packages__ and __extras__ (if defined) to the package object.

    - Automatically checks installed package requirements
      unless `check_requirements` is False.
    - Automatically checks installed package files
      unless `check_packages` is False.
    - Returns the zetup config object.
    """
    try:
        mod = sys.modules[pkgname]
    except KeyError:
        raise ZetupError(
            "Package %s was not found in sys.modules" % repr(pkgname))
    zfg = find_zetup_config(pkgname)
    mod.__version__ = zfg.VERSION
    mod.__requires__ = zfg.REQUIRES
    if check_requirements:
        zfg.REQUIRES.check()
    if zfg.EXTRAS:
        mod.__extras__ = zfg.EXTRAS
    mod.__distribution__ = zfg.DISTRIBUTION.find(os.path.dirname(__file__))
    mod.__description__ = zfg.DESCRIPTION
    mod.__packages__ = zfg.PACKAGES
    if (check_packages
        #TODO: remove (only for backwards compatibility)
        and isinstance(zfg.PACKAGES, Packages)
        ):
        zfg.PACKAGES.check()
    return zfg


class annotate_extra(object):
    def __init__(self, extra=None):
        self.extra = extra

    def __getitem__(self, extra):
        return type(self)(extra)

    def __call__(self, toplevel, pkgname, check_requirements=True):
        if self.extra is None:
            extra = pkgname.rsplit('.', 1)[1]
        else:
            extra = self.extra
        mod = sys.modules[pkgname]
        mod.__version__ = toplevel.__version__
        mod.__requires__ \
            = toplevel.__requires__ + toplevel.__extras__[extra]
        if check_requirements:
            mod.__requires__.check()


annotate_extra = annotate_extra()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.process.scons

call scons with updated PYTHONPATH.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
__all__ = ['scons']

from . import call


def scons(*args, **options):
    """Call external 'scons' command via :func:`zetup.process.call`,
       which updates PYTHONPATH from current ``sys.path``.

    - All lower_case `options` will be passed as ``**kwargs``
      to :func:`zetup.process.call`.
    - All UPPER_CASE `options` will be passed to the 'scons' command.
    """
    kwargs = {}
    for key, value in list(options.items()):
        if not key.isupper():
            kwargs[key] = options.pop(key)
    return call(
      ['scons'] + list(args) + ['%s=%s' % o for o in options.items()],
      **kwargs)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import sys
import re
from textwrap import dedent
from glob import glob

from path import Path
from jinja2 import FileSystemLoader, TemplateNotFound
from jinjatools import Environment

import zetup
from zetup.zetup import Zetup
from zetup.commands import ZetupCommandError


class Loader(FileSystemLoader):
    def __init__(self):
        self.templates_dir = Path(__path__[0]) / 'templates'
        super(Loader, self).__init__(self.templates_dir)

    def get_source(self, env, target):
        source, _, uptodate = super(Loader, self).get_source(
            env, target + '.jinja')
        return source, target, uptodate


class Made(list):
    def __init__(self):
        self.status = 0

    def clean(self):
        for path in self:
            if path.exists():
                print("zetup: Removing auto-generated %s" % path,
                      # don't pollute stdout
                      file=sys.stderr)
                path.remove()
            if path.ext == '.py':
                compiled = []
                path = path.splitext()[0] + '.pyc'
                if path.exists():
                    compiled.append(path)
                compiled += glob(path.dirname() / '__pycache__' /
                                 path.namebase + '.*.pyc')
                for path in map(Path, compiled):
                    print("zetup: Removing compiled %s" % path,
                          # don't pollute stdout
                          file=sys.stderr)
                    path.remove()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.clean()

    def __del__(self):
        self.clean()


class ZetupMakeError(ZetupCommandError):
    def __init__(self, made, message):
        ZetupCommandError.__init__(self, message)
        made.clean()


@Zetup.command()
def make(zfg, args=None, targets=None, force=None, skip_existing=False):
    if args:
        targets = args.targets
        force = args.force
    if force is None:
        force = zfg.FORCE_MAKE
    if not targets:
        raise ZetupCommandError("No targets given. You can 'make all'.")
    env = Environment(loader=Loader())
    if 'all' in targets:
        templates_dir = Path(env.loader.templates_dir)
        targets = [templates_dir.relpathto(tpath).rsplit('.', 1)[0]
                   for tpath in templates_dir.walkfiles()
                   if tpath.endswith('.jinja')]
        skip_existing = True

    made = Made()
    for target in targets:
        if zfg.NO_MAKE and target in zfg.NO_MAKE:
            continue
        if target in ['zetup_config', 'zfg']:
            if zfg.ZETUP_CONFIG_PACKAGE:
                target = '__init__.py'
            elif zfg.ZETUP_CONFIG_MODULE:
                target = 'package/zetup_config.py'
            else:
                continue
        if zfg.PACKAGES:
            target = re.sub(
              '^%s/' % zfg.PACKAGES.main, 'package/', target)
        path = target
        if zfg.PACKAGES:
            path = re.sub(
              '^package', zfg.PACKAGES.main.replace(*'./'), path)
        path = Path(zfg.ZETUP_DIR) / path
        if path.exists():
            if skip_existing:
                print("zetup: NOT generating existing %s" % target,
                      # don't pollute stdout
                      file=sys.stderr)
                continue

            if not force:
                raise ZetupMakeError(
                    made, "Target '%s' already exists. "
                    "Overwrite with -f or --force" % target)
        try:
            template = env.get_template(target)
        except TemplateNotFound:
            raise ZetupMakeError(
                made, "No template for target '%s'." % target)

        print("zetup: Generating %s" % target,
              # don't pollute stdout
              file=sys.stderr)
        text = template.render({
            'zetup_header': dedent("""
            # This file was auto-generated by ZETUP
            #
            # Zimmermann's Extensible Tools for Unified Project setups
            #
            # https://github.com/zimmermanncode/zetup
            """),
            'zetup': zetup,
            'zfg': zfg,
        })
        path.write_text(text.strip())
        if not zfg.KEEP_MADE or target not in zfg.KEEP_MADE:
            made.append(path)

    return made



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.doc

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import sys
from importlib import import_module
from types import ModuleType

import zetup

# zetup.package(__name__, ['AutoDocScopeImporter'])


class AutoDocScopeModule(ModuleType):

    # to be accepted as packages for subscope imports :)
    __path__ = None

    def __init__(self, name, scope):
        ModuleType.__init__(self, scope.__name__)
        modclass = type(self)

        class Module(type(self)):

            def __getattribute__(self, name):
                if name in ['__class__', '__path__']:
                    return modclass.__getattribute__(self, name)
                return getattr(scope, name)

            def __repr__(self):
                return "<%s for %s>" % (modclass.__name__, repr(scope))

        self.__class__ = Module


class AutoDocScopeImporter(object):

    def __init__(self, pkgname):
        package = import_module(pkgname)
        if not isinstance(package, zetup.package):
            raise TypeError("%s is not a zetup.package instance"
                            % repr(package))
        self.package = package

    def find_module(self, name, path):
        if name.startswith(self.package.__name__ + '.'):
            return self

    def load_module(self, name):
        assert name.startswith(self.package.__name__ + '.')
        scope = scopemod = self.package
        for attr in name[len(self.package.__name__) + 1:].split('.'):
            scope = getattr(scope, attr)
            scopemod = AutoDocScopeModule(
                '%s.%s' % (scopemod.__name__, attr), scope)
            scopemod.__path__ = None
            sys.modules[name] = scopemod
        return scope

    # def __repr__(self):
    #     return "<%s for %s>" % (type(self).__name__, repr(self.module))



# ZETUP
#
# Zimmermann's Extensible Tools for Unified Project setups
#
# Copyright (C) 2014-2017 Stefan Zimmermann <user@zimmermann.co>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

"""
Resolver for setup requirements, fetching ``.eggs/`` on demand
"""

import re
import sys

from pkg_resources import (
    get_distribution, working_set, DistributionNotFound, VersionConflict)
from setuptools.dist import Distribution

__all__ = ['resolve']


#: The egg installer
INSTALLER = Distribution().fetch_build_egg


class StdErrWrapper(object):
    """
    For safely redirecting ``stdout`` to ``stderr``

    For example on Windows, directly assigning ``stderr`` to ``stdout`` often
    leads to a detached ``stderr`` buffer in the end
    """

    def __getattr__(self, name):
        return getattr(sys.__stderr__, name)

    def detach(self):
        """
        Don't let ``stderr``'s buffer get stolen
        """
        return self

    def __del__(self):
        """
        Don't let :meth:`.__getattr__` fetch ``stderr``'s ``.__del__``
        """
        pass


def resolve(requirements):
    """
    Make sure that setup `requirements` are always correctly resolved and
    accessible by:

    * Recursively resolving their runtime requirements
    * Moving any installed eggs to the front of ``sys.path``
    * Updating ``pkg_resources.working_set`` accordingly
    """
    # don't pollute stdout! first backup
    __stdout__ = sys.__stdout__
    stdout = sys.stdout
    # then redirect to stderr...
    sys.stdout = sys.__stdout__ = StdErrWrapper()

    def _resolve(requirements, parent=None):
        """
        The actual recursive `requirements` resolver

        `parent` string is used for printing fully qualified requirement
        chains
        """
        for req in requirements:
            qualreq = parent and '%s->%s' % (req, parent) or req
            print("Resolving setup requirement %s:" % qualreq)
            try:
                dist = get_distribution(req)
            except (DistributionNotFound, VersionConflict):
                dist = INSTALLER(req)
                sys.path.insert(0, dist.location)
                working_set.add_entry(dist.location)
            print(repr(dist))
            extras = re.match(r'[^#\[]*\[([^#\]]*)\]', req)
            if extras:
                extras = list(map(str.strip, extras.group(1).split(',')))
            _resolve((str(req).split(';')[0]
                      for req in dist.requires(extras=extras or ())),
                     qualreq)

    _resolve((str(req).split(';')[0] for req in requirements))
    # ... and finally restore stdout
    sys.__stdout__ = __stdout__
    sys.stdout = stdout



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Requirements', 'DistributionNotFound', 'VersionConflict']

import sys
if sys.version_info[0] == 3:
    unicode = str
import re

import pkg_resources
from pkg_resources import \
    parse_requirements, Requirement, get_distribution


class DistributionNotFound(pkg_resources.DistributionNotFound):
    def __init__(self, req, requirer, reason=None):
        super(DistributionNotFound, self).__init__(req, [requirer])
        self.requirer = requirer
        self.reason = reason

    def __str__(self):
        text = "%s needs %s" % (self.requirer, self.req)
        if self.reason:
            text += " (%s)" % self.reason
        return text


class VersionConflict(pkg_resources.VersionConflict):
    def __init__(self, req, found_version, requirer, reason=None):
        super(VersionConflict, self).__init__(
          '%s-%s' % (req.key, found_version), req)
        self.requirer = requirer
        self.reason = reason

    def __str__(self):
        text = "%s needs %s but found %s" % (
          self.requirer, self.req, self.dist)
        if self.reason:
            text += " (%s)" % self.reason
        return text


class Requirements(object):
    """Package requirements manager.
    """
    @staticmethod
    def _parse(text):
        """ Generate parsed requirements from `text`,
            which should contain newline separated requirement specs.

        - Additionally looks for "#import name" comments
          after requirement lines (the actual root module name
          of the required package to use for runtime dependency checks)
          and stores them as .impname attrs on the Requirement instances.
        - Supports #py.. tags at the beginning of lines,
          specifying a python version the requirement applies to.
        """
        for line in text.split('\n'):
            line = line.strip()
            if not line:
                continue
            match = re.match(r'^#py([0-9]+) (.*)$', line)
            if match: #==> only required in given python version
                pyver, line = match.groups()
                #TODO:
                # if len(pyver) > 2:
                if not ('%s%s' % sys.version_info[:2]).startswith(pyver):
                    continue
            try:
                req, impname = line.split('#import')
            except ValueError:
                req = next(parse_requirements(line), None)
                impname = req and req.unsafe_name
            else:
                req = next(parse_requirements(req), None)
            if not req: # maybe a comment line
                continue
            req.impname = impname.strip()
            yield req

    def __init__(self, reqs, zfg=None):
        """Store a list of :class:`pkg_resources.Requirement` instances
           from the given requirement specs
           and additionally store them newline separated
           in the :class:`str` base.

        :param reqs: Either a single string of requirement specs
          or a sequence of strings and/or
          :class:`pkg_resources.Requirement` instances.
        :param zfg: Optional zetup config object
          the requirements are related to.
        """
        if isinstance(reqs, Requirements):
            txt = reqs.txt
            reqlist = list(self._parse(reqs.txt))
        elif isinstance(reqs, (str, unicode)):
            txt = reqs
            reqlist = list(self._parse(reqs))
        else:
            txt = ''
            reqlist = []
            for req in reqs:
                if isinstance(req, (str, unicode)):
                    reqlist.extend(self._parse(req))
                elif isinstance(req, Requirement):
                    reqlist.append(req)
                else:
                    raise TypeError(type(req))
                txt += '\n%s' % req

        self.txt = txt
        self._list = reqlist
        self.zfg = zfg

    def __eq__(self, other):
        return isinstance(other, Requirements) \
          and self._list == other._list

    def __str__(self):
        return '\n'.join(map(str, self))

    def check(self, raise_=True):
        """Check that all requirements are available (importable)
           and their versions match (using modules' __version__ attributes).

        - Fallback to pkg_resources.get_distribution().version
          if no __version__ or is None.

        :param raise_: Raise DistributionNotFound if ImportError
          or VersionConflict if version doesn't match?
          If False just return False in that case.
        """
        requirer = self.zfg and '%s-%s' % (
          self.zfg.NAME, self.zfg.VERSION or '(none)')
        for req in self:
            try:
                mod = __import__(req.impname)
            except ImportError as e:
                if raise_:
                    raise DistributionNotFound(req, requirer,
                      reason="%s: %s" % (type(e).__name__, e))
                return False
            if not req.specs: # No version constraints
                continue
            try:
                version = mod.__version__
                if version is None:
                    # treat the same as if .__version__ not exists
                    # (handle in following except block)
                    # ==> same exception, just different msg
                    raise AttributeError(
                      "module's '__version__' attribute is None")
            except AttributeError as e_no__version__attr:
                try: # try to get version from distribution
                    dist = get_distribution(req.key)
                except pkg_resources.DistributionNotFound as e:
                    if raise_:
                        raise VersionConflict(req, None, requirer,
                          reason="%s: %s. %s: %s" % (
                            e_no__version__attr, mod, type(e).__name__, e))
                    return False
                version = dist.version
            if version not in req:
                if raise_:
                    raise VersionConflict(req, version, requirer)
                return False
        return True

    @property
    def checked(self):
        self.check()
        return self

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, name):
        """Get a requirement by its distribution name.
        """
        for req in self._list:
            if name in [req.key, req.unsafe_name]:
                return req
        raise KeyError(name)

    def __delitem__(self, name):
        """Delete a requirement by its distribution name.
        """
        for req in list(self._list):
            if name in [req.key, req.unsafe_name]:
                return self._list.remove(req)
        raise KeyError(name)

    def __add__(self, reqs):
        """Return a new :class:`Requirements` instance
           with additional `reqs` from string
           or another :class:`Requirements` instance.
        """
        if isinstance(reqs, Requirements):
            reqs = reqs.txt
        return type(self)('%s\n%s' % (
            # For simplicity:
            # Just create explicit #import hints for every requirement:
            # '\n'.join('%s #import %s' % (req, req.impname) for req in self),
            self.txt, reqs), zfg=self.zfg)

    @property
    def py(self):
        return '%s("""\n%s\n""", zfg=zfg)' % (
          type(self).__name__, self.txt) ## '\n'.join(
            ## '%s #import %s' % (req, req.impname) for req in self))

    def __repr__(self):
        return str(self)



"""Test :mod:`zetup.modules`,
   containing the (top-level) package module object wrappers.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from inspect import ismodule
from types import ModuleType

import zetup


def test_zetup_toplevel(zfg, in_repo, in_site_packages):
    """Test :class:`zetup.modules.toplevel` via Zetup's own top-level package.
    """
    assert ismodule(zetup)
    assert isinstance(zetup, zetup.toplevel)

    # is the original zetup module correctly stored?
    assert type(zetup.__module__) is ModuleType
    assert zetup.__module__.__name__ == 'zetup'
    # and does __getattr__ delegation work?
    assert zetup.__file__ is zetup.__module__.__file__
    assert zetup.__path__ is zetup.__module__.__path__

    # the dynamic 'zetup.toplevel' member import above
    # should NOT result in the 'zetup.modules' sub-module
    # being added to the wrapper's __dict__
    assert 'modules' not in zetup.__dict__
    assert 'modules' not in dir(zetup)
    # instead it should end up in the original module's __dict__
    assert 'modules' in zetup.__module__.__dict__



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from glob import glob
from textwrap import dedent

if sys.version_info[0] == 3:
    unicode = str

from .error import ZetupError


class ZetupPackageError(ZetupError):
    """Exception related to zetup package config.
    """
    pass


class ZetupPackageCheckError(ZetupPackageError):
    """Exception related to zetup package config.
    """
    def __init__(self, missing, extra):
        self.missing = list(missing)
        self.extra = list(extra)


class File(str):
    """The name of a package file (source or data)
       with reference to the package and its absolute path.
    """
    def __new__(cls, name, package):
        return str.__new__(cls, name)

    def __init__(self, name, package):
        self.package = package

    @property
    def path(self):
        return os.path.join(self.package.path, self)


class Source(File):
    """The name of a package *.py source file
       with reference to the package and its absolute path.
    """
    pass


class DataFile(File):
    """The name of a package data file
       with reference to the package and its absolute path.
    """
    pass


class Package(str):
    """Package name wrapper with access to
       package path, source files, package data and subpackages,
       and validation features for installed packages.
    """
    def __new__(cls, pkg, **kwargs):
        return str.__new__(cls, pkg)

    def __init__(self, pkg, root=None, path=None, data=None,
                 sources=None, datafiles=None, subpackages=None,
                 zfg=None):
        if not isinstance(pkg, Package):
            pkg = None
        self.root = root or pkg and pkg.root
        self._path = path or pkg and pkg._path
        self.data = data and [os.path.sep.join(d.split('/')) for d in data] \
          or pkg and pkg.data or None
        self._sources = sources and list(sources) \
          or pkg and pkg._sources or None
        self._datafiles = datafiles and list(datafiles) \
          or pkg and pkg._datafiles or None
        if not subpackages and pkg:
            subpackages = pkg._subpackages
        self._subpackages = subpackages and [
          type(self)(spkg, root=self.root) for spkg in subpackages
          ] or None
        self.zfg = zfg or pkg and pkg.zfg

    @property
    def path(self):
        """The absolute path of the package.
        """
        if not os.path.exists(self.root):
            raise RuntimeError(
                "Given root directory for package %s does not exist: %s"
                % (repr(str(self)), self.root))

        if not os.path.isdir(self.root):
            if self.root.endswith('.egg'):
                raise NotImplementedError(
                    "Can't lookup paths of package %s in .egg: %s"
                    % (repr(str(self)), self.root))

            raise RuntimeError(
                "Given root path for package %s is not a directory: %s"
                % (repr(str(self)), self.root))

        return os.path.realpath(os.path.join(self.root or '.',
          self._path or os.path.join(*self.split('.'))))

    def sources(self, force_search=False):
        """Iterates :class:`zetup.package.Source` instances
           of the package's direct source *.py files
           (without sub-package sources).
        """
        if not force_search and self._sources:
            for source in self._sources:
                yield source
            return

        for name in os.listdir(self.path):
            path = os.path.join(self.path, name)
            if os.path.isfile(path) and path.endswith('.py'):
                yield Source(name, package=self)

    def datafiles(self, force_search=False):
        """Iterates :class:`zetup.package.DataFile` instances
           of the package's direct data files
           (without sub-package data files).
        """
        if not force_search and self._datafiles:
            for datafile in self._datafiles:
                yield datafile
            return
        if not self.data:
            return
        for pattern in self.data:
            for path in glob(os.path.join(self.path, pattern)):
                yield DataFile(os.path.relpath(path, self.path),
                               package=self)

    def files(self, force_search=False):
        """Iterates the package's direct sources and data files combined
           (without sub-package files).
        """
        for source in self.sources(force_search=force_search):
            yield source
        for datafile in self.datafiles(force_search=force_search):
            yield datafile

    def subpackages(self, force_search=False):
        """Iterates the package's direct sub-packages as instances of own type
           (without sub-sub-packages).
        """
        if self._subpackages:
            for pkg in self._subpackages:
                yield pkg
            return

        for name in os.listdir(self.path):
            path = os.path.join(self.path, name)
            if os.path.isdir(path) and os.path.isfile(
              os.path.join(path, '__init__.py')):
                yield type(self)('.'.join((self, name)),
                  path=self._path and os.path.join(self._path, name),
                  root=self.root)

    def walk(self):
        """Iterates the package's sub-packages recursively.
        """
        for pkg in self.subpackages():
            yield pkg
            for subpkg in pkg.walk():
                yield subpkg

    def issubpackage(self, parent, recursive=False):
        pkg = self.split('.')
        parent = parent.split('.')
        if len(pkg) <= len(parent):
            return False
        if not recursive and (len(pkg) - len(parent) > 1):
            return False
        return parent == pkg[:len(parent)]

    def __getitem__(self, name):
        """Get a subpackge by its relative name.
        """
        for pkg in self.subpackages():
            if pkg == '.'.join((self, name)):
                return pkg
        raise KeyError("%s has no sub-package named %s"
                       % (repr(self), repr(name)))

    def __getattr__(self, name):
        """Fast interactive way for the shell
           to get a subpackge by its relative name.
        """
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
              "%s has no attribute or sub-package named %s"
              % (repr(self), repr(name)))

    def check(self, raise_=True):
        #TODO
        if self.root.endswith('.egg') and not os.path.isdir(self.root):
            return True

        if self._sources is not None:
            expected = set(self._sources)
            found = set(self.sources(force_search=True))
            diff = expected - found
            if diff:
                if raise_:
                    raise RuntimeError(diff)
                return False

        if self._datafiles is not None:
            expected = set(self._datafiles)
            found = set(self.datafiles(force_search=True))
            diff = expected - found
            if diff:
                if raise_:
                    raise RuntimeError(diff)
                return False

        if self._subpackages is not None:
            expected = set(self._subpackages)
            found = set(self.subpackages(force_search=True))
            diff = expected - found
            if diff:
                if raise_:
                    raise RuntimeError(diff)
                return False

        return True

    @property
    def py(self):
        """Generate Python code for zetup config module.
        """
        return dedent("""
          %s(%s,
            sources=[
              %s
              ],
            subpackages=[
              %s
              ],
            )
          """) % (type(self).__name__, repr(str(self)),
                  ",\n    ".join(repr(os.path.basename(src)) for src in self.sources()),
                  ",\n    ".join(pkg.py for pkg in self.subpackages()))


class Packages(object):
    def _parse(self, text, root=None):
        packages = []
        for line in map(str.strip, text.split('\n')):
            if not line:
                continue
            try:
                pkg, data = map(str.strip, line.split('+'))
            except ValueError:
                pkg, data = line, None
            else:
                data = data.split()
            try:
                pkg, path = map(str.strip, pkg.split(':'))
            except ValueError:
                path = None
            packages.append(Package(pkg, root=root, path=path, data=data))
        toplevel = list(packages)
        for pkg in sorted(packages):
            for item in packages:
                if pkg.issubpackage(item):
                    if item._subpackages:
                        item._subpackages.append(pkg)
                    else:
                        item._subpackages = [pkg]
                        toplevel.remove(pkg)
                    break
        return toplevel

    def __init__(self, text_or_toplevel, root=None, zfg=None):
        self.root = root
        self.zfg = zfg
        if isinstance(text_or_toplevel, (str, unicode)):
            self.toplevel = self._parse(text_or_toplevel, root=root)
        else:
            self.toplevel = [Package(pkg, root=root)
                             for pkg in text_or_toplevel]

    def __iter__(self):
        """Iterate all toplevel and sub-packages.
        """
        for pkg in self.toplevel:
            yield pkg
            for subpkg in pkg.walk():
                yield subpkg

    def __len__(self):
        return len(list(iter(self)))

    def __getitem__(self, name):
        for pkg in self:
            if pkg == name:
                return pkg
        raise KeyError("No package named %s" % repr(name))

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
              "%s instance has no attribute or package named %s"
              % (type(self), repr(name)))

    def __bool__(self):
        """Any packages?
        """
        return bool(self.toplevel)

    #PY2
    def __nonzero__(self):
        return self.__bool__()

    @property
    def main(self):
        return next(iter(self))

    def check(self, raise_=True):
        #TODO
        if self.root.endswith('.egg') and not os.path.isdir(self.root):
            return True

        for pkg in self:
            pkg.check(raise_=raise_)

    @property
    def checked(self):
        self.check()
        return self

    @property
    def py(self):
        """Generate Python code for zetup config module.
        """
        if self.zfg:
            zfg_modname = self.zfg.ZETUP_CONFIG_MODULE
            root_code = "os.path.realpath(__file__)"
            for _ in range(zfg_modname.count('.') + 1):
                root_code = "os.path.dirname(%s)" % root_code
        else:
            root_code = 'None'
        return "%s([\n%s\n  ], root=%s)" % (type(self).__name__,
          ",\n  ".join(pkg.py for pkg in self.toplevel),
          root_code)

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, repr(self.toplevel))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Notebook']

# TODO: not working because module is imported
# before main zetup module is annotated
#
# import zetup
#
# __requires__ = zetup.__requires__ + zetup.__extras__['notebook']

import re
from inspect import getmembers

try:
    #TODO: zetup.Path base
    from path import Path as base
except ImportError:
    # just reduce features
    base = str

from zetup.object import object


def check_requirements(raise_=True):
    """Check requirements for full :class:`zetup.Notebook` features.
    """
    import zetup
    return zetup.__extras__['notebook'].check(raise_=raise_)


class Notebook(object, base):

    def __dir__(self):
        """Return names of dynamically generatable
           ``.to_...()`` converter methods
           with ``nbconvert.export_...()`` delegation.
        """
        if not check_requirements(raise_=False):
            return []
        # import on demand to be not required for module import
        import nbconvert
        return super(Notebook, self).__dir__() + [
            'to_' + name for name in nbconvert.get_export_names()]

    def __getattr__(self, name):
        """Dynamically generate ``.to_...()`` converter methods
           with ``nbconvert.export_...()`` delegation.
        """
        if not name.startswith('to_'):
            raise AttributeError(
              "%s instance has no attribute %s" % (type(self), repr(name)))

        check_requirements()
        # import on demand to be not required for module import
        import nbconvert
        try:
            exportercls = nbconvert.get_exporter(name.split('_', 1)[1])
        except ValueError as e:
            raise AttributeError(
              "%s instance has no converter method %s (%s: %s)"
              % (type(self), repr(name), type(e).__name__, e))

        def method():
            result = exportercls().from_filename(self)
            return result[0] # only the actual text

        method.__name__ = name
        return method

    def to_bitbucket_rst(self):
        """Export the notebook to Bitbucket optimized reST.

        - Prepends 'rst-header-' to link targets.
          ('#some-section' --> '#rst-header-some-section')
        """
        check_requirements()
        # import on demand to be not required for module import
        import nbconvert
        from .jinja import FILTERS, LOADER

        exporter = nbconvert.RSTExporter()
        exporter.filters = FILTERS
        exporter.extra_loaders = [LOADER]
        # not a real file. will be loaded from string by LOADER
        exporter.template_file = 'bitbucket_rst'
        rst = exporter.from_filename(self)[0]
        # bitbucket_rst template puts code cell input and output
        #  in single blocks, but can't prevent empty lines in between
        #==> Remove them:
        return re.sub(
          # .. code:: python
          #
          #     >>> input
          #
          #     output
          r'(\n    >>> .+\n)\s*\n(    [^>])',
          # .. code:: python
          #
          #     >>> input
          #     output
          r'\1\2',
          rst)

    def to_github_markdown(self):
        """Export the notebook to Github optimized Markdown.

        - Removes header enumeration prefixes from link targets.
          ('#1.-section-one' --> '#section-one')
        """
        check_requirements()
        # import on demand to be not required for module import
        import nbconvert
        from .jinja import FILTERS, LOADER

        exporter = nbconvert.MarkdownExporter()
        exporter.filters = FILTERS
        exporter.extra_loaders = [LOADER]
        # not a real file. will be loaded from string by LOADER
        exporter.template_file = 'github_markdown'
        markdown = exporter.from_filename(self)[0]
        # github_markdown template puts code cell input and output
        #  in single blocks, but can't prevent empty lines in between
        #==> Remove them:
        markdown = re.sub(
          # ```python
          # >>> input
          #
          # output
          # ```
          r'(\n>>> .+\n)\s*\n([^>`])',
          # ```python
          # >>> input
          # output
          # ```
          r'\1\2',
          markdown)
        # And also remove newlines after ```python and before ```:
        markdown = re.sub(r'(\n```python.*\n)\s*', r'\1', markdown)
        markdown = re.sub(r'\n\s*(```\s*\n)', r'\n\1', markdown)
        return markdown



# This file was auto-generated by ZETUP
#
# Zimmermann's Extensible Tools for Unified Project setups
#
# https://github.com/zimmermanncode/zetup


from __future__ import absolute_import

import sys
import os
from collections import OrderedDict

from zetup.version import Version
from zetup.dist import Distribution
from zetup.requires import Requirements
from zetup.extras import Extras
from zetup.package import Packages, Package
from zetup.notebook import Notebook



zfg = sys.modules[__name__]

NAME = 'zetup'

LICENSE = 'LGPLv3'

TITLE = 'zetup'

PYTHON = ['2.7', '3.3', '3.4', '3.5', '3.6']

MODULES = []

FORCE_MAKE = True

KEYWORDS = ['zetup', 'project', 'tools', 'config', 'ini', 'setup', 'requirements', 'extras', 'packages', 'package', 'toplevel', 'object', 'meta', 'subprocess', 'process', 'popen', 'call', 'python3', 'python3']

PACKAGES = Packages([

Package('zetup',
  sources=[
    '__init__.py',
    '__main__.py',
    'annotate.py',
    'classpackage.py',
    'conda.py',
    'config.py',
    'dist.py',
    'error.py',
    'extras.py',
    'modules.py',
    'object.py',
    'package.py',
    'requires.py',
    'resolve.py',
    'script.py',
    'test_classpackage.py',
    'version.py',
    'zetup.py'
    ],
  subpackages=[
    
Package('zetup.commands',
  sources=[
    '__init__.py',
    'command.py',
    'conda.py',
    'del_.py',
    'dev.py',
    'error.py',
    'pytest.py',
    'run.py',
    'test.py',
    'tox.py'
    ],
  subpackages=[
    
Package('zetup.commands.make',
  sources=[
    '__init__.py'
    ],
  subpackages=[
    
    ],
  )

    ],
  )
,
    
Package('zetup.doc',
  sources=[
    '__init__.py'
    ],
  subpackages=[
    
    ],
  )
,
    
Package('zetup.notebook',
  sources=[
    '__init__.py',
    'jinja.py'
    ],
  subpackages=[
    
    ],
  )
,
    
Package('zetup.process',
  sources=[
    '__init__.py',
    'scons.py'
    ],
  subpackages=[
    
    ],
  )

    ],
  )

  ], root=os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

EMAIL = 'user@zimmermann.co'

AUTHOR = 'Stefan Zimmermann'

EXTRAS = Extras([
('commands', Requirements("""
path.py >= 10.3 #import path
jinjatools >= 0.1.7

""")),
('notebook', Requirements("""
path.py >= 10.3 #import path
nbconvert >= 5.2

""")),
('pytest', Requirements("""
pytest

"""))
], zfg=zfg)

REQUIRES = Requirements("""
# setuptools
setuptools_scm >= 1.15

""", zfg=zfg)

URL = 'https://github.com/zimmermanncode/zetup'

SETUP_HOOKS = []

VERSION = Version('0.2.42')

KEEP_MADE = []

NO_MAKE = ['setup.py', 'tox.ini']

TEST_COMMANDS = ['py.test -v test']

DESCRIPTION = "Zimmermann's Extensible Tools for Unified Project_setups"

SETUP_REQUIRES = None

in_repo = True

SCRIPTS = {'zetup': 'zetup.script:run'}

SETUP_KEYWORDS = {'use_zetup': 'zetup:setup_entry_point'}

DISTRIBUTION = Distribution(zfg)

CLASSIFIERS = ['Development Status :: 3 - Alpha', 'Intended Audience :: Developers', 'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)', 'Operating System :: OS Independent', 'Topic :: Software Development', 'Topic :: Utilities', 'Programming Language :: Python', 'Programming Language :: Python :: 2.7', 'Programming Language :: Python :: 3.3', 'Programming Language :: Python :: 3.4', 'Programming Language :: Python :: 3.5', 'Programming Language :: Python :: 3.6']


# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.test_classpackage

pytest unit tests for :class:`zetup.classpackage`.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

import zetup


def test_bases():
    for base in [zetup.object, zetup.package]:
        assert issubclass(zetup.classpackage, base)
    for nobase in [zetup.toplevel]:
        assert not issubclass(zetup.classpackage, nobase)


def test_metabases():
    assert type(zetup.classpackage) is zetup.meta


def test__module__():
    assert zetup.classpackage.__module__ == 'zetup'


def test_members():
    # should not contain more members than basic zetup.package
    assert dir(zetup.classpackage) == dir(zetup.package)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

import re
import os

from path import Path

from zetup.process import call
from zetup.zetup import Zetup


@Zetup.command(depends=['VERSION', 'setup.py', '__init__.py'])
def conda(zfg, args):
    """The actual conda command action called by Command base.
    """
    import yaml

    metadir = Path('.conda')
    metadir.mkdir_p()
    metafile = metadir / 'meta.yaml'
    buildfile = metadir / 'build.sh'

    def conda_req(req):
        """conda wants space between requirement name and version specs.
        """
        return re.sub(r'([=<>]+)', r' \1', str(req))

    requirements = list(map(conda_req, zfg.REQUIRES))
    # Also add all extra requirements
    #  (conda doesn't seem to have such an extra features management):
    for extra in zfg.EXTRAS.values():
        requirements.extend(map(conda_req, extra))

    meta = { # to be dumped to meta.yaml
      'package': {
        'name': zfg.NAME,
        'version': str(zfg.VERSION),
        },
      'source': {
        'fn': '%s-%s.tar.gz' % (zfg.NAME, zfg.VERSION),
        # The absolute path to the sdist in dist/
        'url': 'file://%s' % os.path.realpath(os.path.join(
          'dist', '%s-%s.tar.gz' % (zfg.NAME, zfg.VERSION)))
        },
      'requirements': {
        'build': [
          'python',
          'pyyaml',
          ] + requirements,
        'run': [
          'python',
          ] + requirements,
        },
      'about': {
        'home': zfg.URL,
        'summary': zfg.DESCRIPTION,
        },
      }
    with open(metafile, 'w') as f:
        yaml.dump(meta, f, default_flow_style=False)

    with open(buildfile, 'w') as f:
        f.write('#!/bin/bash'
                '\n\n'
                '$PYTHON setup.py install'
                '\n')

    return call(['conda', 'build', metadir], env=os.environ)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.process

subprocess wrappers for calling external programs with updated PYTHONPATH
and better Windows support by implicitly calling .bat and .cmd scripts
without explicitly specified file extension.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
__all__ = ['call', 'Popen']

import sys
import os
from itertools import chain
import subprocess
from subprocess import PIPE

if sys.version_info[0] == 3:
    unicode = str


# define a platform-dependent `_command` helper,
# which converts a `command` arg for subprocess.Popen or .call
# to support Windows .bat and .cmd scripts without extension
if sys.platform.startswith('win'):

    def _where(command):
        """Try to find the executable file path of given `command`
        using ``where.exe``.
        """
        if os.path.isabs(command):
            return command

        out = subprocess.Popen(
            ['where.exe', command], stdout=PIPE, stderr=PIPE,
            universal_newlines=True
        ).communicate()
        if not out[0]:
            raise FileNotFoundError("Can't find executable %s"
                                    % repr(command))
        return out[0].split('\n', 1)[0].strip()

    try:
        # find executable path with win32api.FindExecutable()
        # if pywin32 is installed
        import win32api
        import pywintypes
    except ImportError:
        # otherwise find executable path with where.exe
        def _command(command, kwargs):
            if kwargs.get('shell'):
                return command
            if isinstance(command, (str, unicode)):
                try:
                    command, args = command.split(None, 1)
                except ValueError:
                    args = ''
                path = _where(command)
                return args and " ".join((path, args)) or path
            # else command is sequence
            args = iter(command)
            command = next(args)
            path = _where(command)
            return chain([path], args)
    else:
        # have pywin32
        def _find(command):
            """Try to find the executable file path of given `command`
               using ``win32api.FindExecutable()``.
            """
            try:
                status, path = win32api.FindExecutable(command)
            except pywintypes.error:
                # try again with where.exe
                # (FindExecutable fails for commands with dots)
                return _where(command)
            return path

        def _command(command, kwargs):
            if kwargs.get('shell'):
                return command
            if isinstance(command, (str, unicode)):
                try:
                    command, args = command.split(None, 1)
                except ValueError:
                    args = ''
                path = _find(command)
                return args and " ".join((path, args)) or path
            # else command is sequence
            args = iter(command)
            command = next(args)
            path = _find(command)
            return chain([path], args)

else: # no Windows ==> just pass through
    def _command(command, kwargs):
        return command


def _prepare_kwargs(kwargs):
    """Prepare (manipulate) the given `kwargs` dict for passing to
       :func:`subprocess.call` or :class:`subprocess.Popen`
       by resolving any zetup-specific stuff.
    """
    env = kwargs.get('env')
    if env is None:
        env = kwargs['env'] = dict(os.environ,
          PYTHONPATH=os.pathsep.join(sys.path))

    env.update(kwargs.pop('env_update', {}))
    for key, value in kwargs.pop('env_defaults', {}).items():
        env.setdefault(key, value)


class Popen(subprocess.Popen):
    """Wrapper for :func:`subprocess.Popen`,
       which updates PYTHONPATH from current ``sys.path``.

    - Supports same `kwargs` as ``subprocess`` implementation.
    - Setting ``env=`` overrides whole environment;
      to keep PYTHONPATH update, use ``env_update=`` and ``env_defaults=``
      with dicts containing only the variables to change.
    - On Windows, supports running scripts without explicitly adding
      ``'.bat'`` or ``'.cmd'`` extensions.
      **pywin32** should be installed for better performance.
    """
    def __init__(self, command, **kwargs):
        _prepare_kwargs(kwargs)
        command = _command(command, kwargs)
        subprocess.Popen.__init__(self, command, **kwargs)


def call(command, **kwargs):
    _prepare_kwargs(kwargs)
    command = _command(command, kwargs)
    return subprocess.call(command, **kwargs)

call.__doc__ = Popen.__doc__.replace('Popen', 'call')



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.conda

Provides a convenience ``conda`` interface to external conda executable.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

import os
from subprocess import PIPE
import json

from zetup.process import Popen, call
from zetup.object import object

__all__ = ['conda']


class Conda(object):
    """Creates an interface to external conda executable.
    """
    def __init__(self, command=None):
        """Implicitly use the optional conda `command`
           as first conda argument.
        """
        self.command = command

    def __call__(self, *args, **options):
        """Runs external conda executable with the given `args`.

        - Automatically adds ``--json`` arg
          and returns pythonized JSON output.
        - If called with ``json=False``,
          stdout is not captured and status code is returned.
        """
        args = list(args)
        if self.command:
            args.insert(0, self.command)
        if options.get('json', True):
            return json.loads(Popen(
                ['conda'] + args + ['--json'], env=os.environ,
                stdout=PIPE, universal_newlines=True
            ).communicate()[0])
        return call(['conda'] + args, env=os.environ)

    def __getattr__(self, command):
        """Get an interface for a specific conda `command`.
        """
        if command.startswith('_'):
            raise AttributeError("%s has no attribute %s"
                                 % (repr(self), repr(command)))
        return type(self)(command)


# the actual conda interface
conda = Conda()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from functools import partial

__all__ = ['command']


COMMANDS = {}


def command(func=None, name=None):
    """Decorator for registering basic (non-project-bound) commands.
    """
    if func is None:
        if name:
            return partial(command, name=name)
        raise TypeError(
            "zetup.command.command() decorator "
            "takes at least a function or a name argument.")
    COMMANDS[name or func.__name__] = func
    return func



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Extras']

from itertools import chain
from collections import OrderedDict

from .requires import Requirements


class Extras(OrderedDict):
    """Package extra features/requirements manager.

    - Stores :class:`Requirements` instances by extra feature name keys.
    - Provides an implicit 'all' key, returning a dynamically created
      combined :class:`Requirements` instance with all extra requirements.
    """
    def __init__(self, mapping=(), zfg=None):
        self.zfg = zfg
        super(Extras, self).__init__(mapping)

    def __setitem__(self, name, text):
        reqs = Requirements(text, zfg=self.zfg)
        super(Extras, self).__setitem__(name, reqs)

    def __getitem__(self, name):
        if name == 'all':
            return Requirements(chain(*self.values()), zfg=self.zfg)
        return super(Extras, self).__getitem__(name)

    @property
    def py(self):
        return '%s([\n%s\n], zfg=zfg)' % (type(self).__name__, ',\n'.join(
          '(%s, %s("""\n%s\n"""))' % (
            repr(name), type(reqs).__name__, reqs.txt) ## '\n'.join(
              ## '%s #import %s' % (req, req.impname) for req in reqs))
          for name, reqs in self.items()))

    def __repr__(self):
        return "\n\n".join((
          "[%s]\n" % key + '\n'.join(map(str, reqs))
            for key, reqs in self.items() if key != 'all'))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Version']

import sys
if sys.version_info[0] == 3:
    unicode = str

from pkg_resources import parse_version


class Version(str):
    """Manage and compare version strings
       using :func:`pkg_resources.parse_version`.
    """
    @staticmethod
    def _parsed(value):
        """Parse a version `value` if needed (if simple string).
        """
        if isinstance(value, (str, unicode)):
            value = parse_version(value)
        return value

    @property
    def parsed(self):
        """The version string as parsed version tuple.
        """
        return parse_version(self)

    # Need to override all the compare methods
    #  (would otherwise be taken from str base)...
    def __eq__(self, other):
        return self.parsed == self._parsed(other)

    def __ne__(self, other):
        return self.parsed != self._parsed(other)

    def __lt__(self, other):
        return self.parsed < self._parsed(other)

    def __le__(self, other):
        return self.parsed <= self._parsed(other)

    def __gt__(self, other):
        return self.parsed > self._parsed(other)

    def __ge__(self, other):
        return self.parsed >= self._parsed(other)

    @property
    def py(self):
        return '%s(%s)' % (type(self).__name__, repr(str(self)))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.commands.test

Defines ``zetup test`` command.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import os

from zetup.process import call
from zetup.zetup import Zetup

from zetup.commands.dev import dev


@Zetup.command()
def test(zfg, args=None):
    """Run configured ``test commands``.
    """
    dev(zfg)  # first (re)install project in develop mode
    for command in zfg.TEST_COMMANDS:
        print("zetup: Running %s" % repr(command))
        status = call(command, shell=True, env=os.environ)
        if status:  # ==> error
            return status



"""Test the zetup.notebook.Notebook class.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from inspect import getmembers

import nbconvert

import pytest


class TestNotebook(object):
    """Test zetup.Notebook features using zetup's own README notebook.
    """
    def test_converters(self, notebook, in_site_packages):
        if in_site_packages:
            # notebooks aren't installed
            assert notebook is None
            return
        # check if all nbconvert exporters are exposed to the notebook
        # instance as to_...() methods
        for name in nbconvert.get_export_names():
            converter = getattr(notebook, 'to_' + name)
            assert callable(converter)
        # and if the conversion actually works by testing with markdown export
        assert nbconvert.MarkdownExporter().from_filename(notebook)[0] \
            == notebook.to_markdown()

    def test_getattr(self, notebook, in_site_packages):
        if in_site_packages:
            # notebooks aren't installed
            assert notebook is None
            return
        # check correct AttributeError messages
        # (if a non-existing attribute starts with 'to_',
        #  __getattr__ should report that such a 'converter'
        #  instead of 'attribute' does not exist)
        with pytest.raises(AttributeError) as exc:
            notebook.not_defined
        assert all(text in str(exc.value) for text in [
          "has no attribute", "not_defined"])
        assert "converter" not in str(exc.value)
        with pytest.raises(AttributeError) as exc:
            notebook.to_not_defined()
        assert all(text in str(exc.value) for text in [
          "has no converter", "to_not_defined"])



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.modules

Module object wrappers for packages, top-level packages,
and top-level packages for extra features.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

__all__ = ['package', 'toplevel']

import sys
from warnings import warn
from importlib import import_module
from inspect import ismodule
from types import ModuleType
from itertools import chain

import zetup
from .object import object, meta
from .annotate import annotate, annotate_extra
from .doc import AutoDocScopeModule


class deprecated(str):
    """Simple ``str`` class wrapper for marking a
       :class:`zetup.package` api member (alias) name as deprecated.
    """
    def __repr__(self):
        return "deprecated(%s)" % str.__repr__(self)


class package(ModuleType, object):
    """Package module object wrapper
       for clean dynamic API import from sub-modules.
    """
    __module__ = __package__

    def __init__(
            self, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            __getitem__=None, __iter__=None, __call__=None
    ):
        """Wraps a package module given by its `name`.

        - Original package module object is replaced in ``sys.modules``
          and stored in :attr:``.__module__``.
        - Optional `__all__` list defines the package API.
        - Optional `aliases` and `deprecated_aliases`
          map alternative names to API names.
        - `__getitem__`, `__iter__`, and `__call__` features
          can be added to the package wrapper
          by providing handler functions or other callable objects
          (which are not called with a 'self' argument).
        """
        mod = sys.modules[name]
        ModuleType.__init__(self, name, mod.__doc__)
        self.__name__ = name
        self.__module__ = mod
        sys.modules[name] = self
        self.__dict__['__all__'] = api \
            = dict.fromkeys(__all__) if __all__ is not None else {}
        if aliases is not None:
            api.update(aliases)
        if deprecated_aliases is not None:
            api.update((deprecated(alias), name)
                       for alias, name in dict(deprecated_aliases).items())
        # if api is not None:
        #     for submodname, members in dict(__all__).items():
        #         self.__dict__['__all__'].update(
        #             (name, submodname) for name in members)
        cls = type(self)
        cls.__getitem__.funcs[self] = __getitem__
        cls.__iter__.funcs[self] = __iter__
        cls.__call__.funcs[self] = __call__

    def __getitem__(self, key):
        cls = type(self)
        func = cls.__getitem__.funcs.get(self)
        if func is None:
            raise TypeError(
                "%s is not subscriptable. "
                "Instantiate %s with __getitem__=<func> to change that."
                % (repr(self), repr(cls)))
        return func(key)

    __getitem__.funcs = {}

    def __iter__(self):
        cls = type(self)
        func = cls.__iter__.funcs.get(self)
        if func is None:
            raise TypeError(
                "%s is not iterable. "
                "Instantiate %s with __iter__=<func> to change that."
                % (repr(self), repr(cls)))
        return func()

    __iter__.funcs = {}

    def __call__(self, *args, **kwargs):
        cls = type(self)
        func = cls.__call__.funcs.get(self)
        if func is None:
            raise TypeError(
                "%s is not callable. "
                "Instantiate %s with __call__=<func> to change that."
                % (repr(self), repr(cls)))
        return func(*args, **kwargs)

    __call__.funcs = {}

    @property
    def __all__(self):
        """Get API member name list (without deprecated aliases).
        """
        return list(set(chain(
            getattr(self.__module__, '__all__', ()),
            (name for name in self.__dict__['__all__']
             if not isinstance(name, deprecated)))))

    def __setattr__(self, name, value):
        """
        Prevent submodules from being added as attributes to avoid
        unnecessary ``dir()`` pollution.
        """
        if isinstance(value, AutoDocScopeModule) or ismodule(value) and (
                value.__name__ == '%s.%s' % (self.__name__, name)
                and not isinstance(value, package)
        ):
            return
        from .classpackage import classpackage
        if isinstance(value, classpackage):
            value = getattr(value, name)
        object.__setattr__(self, name, value)

    def __getattribute__(self, name):
        """Dynamically access API from wrapped module
           or import extra API members.
        """
        if name.startswith('__'):
            try:
                return object.__getattribute__(self, name)
            except AttributeError:
                pass

        # check if name is defined as alias
        realname = self.__dict__['__all__'].get(name)
        if realname is not None:
            if isinstance(name, deprecated):
                warn("%s.%s is deprecated in favor of %s.%s"
                     % (self.__name__, name, self.__name__, realname),
                     DeprecationWarning)
            name = realname

        try:  # first try to get attr from wrapped original module
            return getattr(self.__module__, name)
        except AttributeError:
            try: # then from wrapper module
                obj = self.__dict__[name]
            except KeyError:
                if name in getattr(self.__module__, '__all__', ()):
                    raise AttributeError(
                        "%s has no attribute %s although listed in __all__"
                        % (repr(self.__module__), repr(name)))
            try: # and finally try to find a matching submodule
                obj = import_module('%s.%s' % (self.__name__, name))
            except ImportError as exc:
                raise AttributeError("%s has no attribute %s"
                                     % (repr(self), repr(name)))

        if isinstance(obj, zetup.classpackage):
            classobj = getattr(obj, name)
            setattr(self, name, classobj)
            return classobj
        return obj

    def __dir__(self):
        """Additionally get all API member names.
        """
        def exclude():
            """Get names of submodules which were implicitly added
               to this package wrapper's ``__dict__``
               by ``from .submodule import ...`` statements in wrapped module
               following the wrapper instantiation.
            """
            for name, obj in self.__dict__.items():
                if name.startswith('__'):
                    continue
                if ismodule(obj) and not isinstance(obj, package):
                    yield name

        return list(chain(
            # (name for name in self.__module__.__dict__ if name.startswith('__')),
            set(object.__dir__(self)).difference(exclude()),
            self.__all__))

    def __repr__(self):
        """Create module-style representation.
        """
        return "<%s %s from %s>" % (
            type(self).__name__, repr(self.__name__),
            repr(self.__module__.__file__))


class toplevel(package):
    """Special top-level package module object wrapper
       for clean dynamic API import from sub-modules
       and automatic application of func:`zetup.annotate`.
    """
    __module__ = __package__

    def __init__(
            self, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            check_requirements=True, check_packages=True,
            __getitem__=None, __iter__=None, __call__=None
    ):
        """Wrap top-level package module given by its `name`
           and `api` member list.

        - See :class:`zetup.package`
          for details about defining the package API and special features.
        - See :func:`zetup.annotate` for details about the check options.
        """
        super(toplevel, self).__init__(
            name, __all__,
            aliases=aliases, deprecated_aliases=deprecated_aliases,
            __getitem__=__getitem__, __iter__=__iter__, __call__=__call__
        )
        zfg = annotate(name, check_requirements=check_requirements,
                       check_packages=check_packages)
        self.__package__ = pkg = zfg.PACKAGES[name]
        pkg.zetup_config = zfg


class extra_toplevel_meta(meta):
    extra = None

    def __getitem__(cls, extra):
        mcs = type(cls)
        meta = type('%s[%s]' % (mcs.__name__, extra), (mcs, ), {
            'extra': extra,
        })
        return meta('%s[%s]' % (cls.__name__, extra), (cls, ), {})


class extra_toplevel(
        # PY2/3 compatible way to assign metaclass
        extra_toplevel_meta('extra_toplevel_base', (package, ), {})
):
    """Special extra feature top-level package module object wrapper
       for clean dynamic API import from sub-modules
       and automatic application of func:`zetup.annotate_extra`.
    """
    __module__ = __package__

    def __init__(
            self, toplevel, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            check_requirements=True,
            __getitem__=None, __iter__=None, __call__=None
    ):
        """Wrap top-level package module given by its `name`
           and `api` member list.

        - See :class:`zetup.package`
          for details about defining the package API and special features.
        - See :func:`zetup.annotate_extra` for details about the check option.
        """
        super(extra_toplevel, self).__init__(
            name, __all__,
            aliases=aliases, deprecated_aliases=deprecated_aliases,
            __getitem__=__getitem__, __iter__=__iter__, __call__=__call__
        )
        extra = type(self).extra
        annotate_extra[extra](
            toplevel, name, check_requirements=check_requirements)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.classpackage

Defines :class:`zetup.classpackage`

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

__all__ = ['classpackage']

from importlib import import_module

from .modules import package


class classpackage(package):
    """Sub-package module wrapper, auto-importing a class definition
    from that sub-package and supporting class member definitions
    in separate sub-modules, if:

    * The class is named after the sub-package.
    * The class is a sub-class of :class:`zetup.object`.
    * The sub-package name is defined as API member of parent
      :class:`zetup.package` instance.
    * Sub-module names are listed via ``membermodules=`` argument
      and members in those sub-modules are decorated
      with class method :func:`zetup.object.member`.

    **package/__init__.py** ::

       import zetup
       zetup.package(__name__, ['Class', ...])

    **package/Class/__init__.py** ::

       import zetup
       zetup.classpackage(__name__, membermodules=['methods', ...])

       class Class:
           def __init__(self, ...):
               ...
           ...

    **package/Class/methods.py** ::

       from . import Class

       @Class.member
       def method(self, ...):
           ...

    >>> import package
    >>> package.Class
    <class package.Class>
    >>> package.Class.method
    <function package.Class.method>
    """
    __module__ = __package__

    def __init__(self, pkgname, membermodules=None):
        """Created with ``__name__`` of the subpackage defining the class
        and the optional list of `membermodules` to automatically import
        (sub-modules defining additional class members).
        """
        parentpkgname, classname = pkgname.rsplit('.', 1)
        super(classpackage, self).__init__(pkgname, [classname])

        pkgcls = type(self)
        pkgdict = self.__dict__

        def load_class():
            """Get the actual class object from this class package
            and automatically import all defined ``membermodules``.
            """
            classobj = pkgcls.__getattribute__(self, classname)
            classobj.__module__ = parentpkgname
            if membermodules is not None:
                for modname in membermodules:
                    mod = import_module('%s.%s' % (pkgname, modname))
            return classobj

        class package(type(self)):
            """Help tools like sphinx ``.. autoclass::`` doc generator
            by delegating (almost) all attributes to the actual class object.

            * ``.. autoclass:: package.Class`` first looks up
              ``package.Class`` in imported modules and therefore
              doesn't get the actual class object itself.
            """
            def __getattribute__(self, name):
                if name == classname:
                    # get the actual class object
                    try:
                        return pkgdict[classname]
                    except KeyError:
                        # only load once
                        classobj = pkgdict[classname] = load_class()
                        return classobj
                # only take these few special attributes
                # from the classpackage instance itself
                if name in [
                        '__name__', '__all__',  '__module__', '__path__',
                        '__file__', '__class__', '__dict__',
                ]:
                    return pkgcls.__getattribute__(self, name)
                # and take all other attributes from the class object
                # (by first recursively calling this __getattribute__)
                return getattr(getattr(self, classname), name)

            def __repr__(self):
                return "<%s for %s from %s>" % (
                    pkgcls.__name__, repr(getattr(self, classname)),
                    repr(self.__file__))

        self.__class__ = package



"""Common fixtures for zetup tests.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from inspect import ismodule

from path import Path

import zetup
from zetup import find_zetup_config

import pytest


@pytest.fixture(scope='module')
def zfg():
    """Get zetup's own config.
    """
    return find_zetup_config('zetup')


@pytest.fixture(scope='module')
def in_site_packages():
    """Is zetup imported from repository?
    """
    return ismodule(zfg())


@pytest.fixture(scope='module')
def in_repo(in_site_packages):
    """Is zetup imported from repo?
    """
    return not in_site_packages


@pytest.fixture(scope='module')
def pkg_path():
    """The dir path of the zetup package.
    """
    return Path(zetup.__file__).realpath().dirname()


@pytest.fixture(scope='module')
def zfg_path(zfg, in_repo, in_site_packages):
    """The dir path of the zetup config.
    """
    if in_repo:
        return Path(zfg.ZETUP_DIR).realpath()
    if in_site_packages:
        return Path(zfg.__file__).realpath().dirname()


@pytest.fixture(scope='module')
def notebook(zfg, in_repo, in_site_packages):
    """zetup's own README.ipynb as :class:`zetup.Notebook` instance
       from zetup's own config.
    """
    if in_site_packages:
        # notebooks aren't installed
        assert 'NOTEBOOKS' not in dir(zfg)
        return None

    return zfg.NOTEBOOKS['README']



from __future__ import print_function

import sys
import os

# from setuptools import Distribution
from pkg_resources import get_distribution, working_set, VersionConflict


def samefile(path, other):
    """Workaround for missing ``os.path.samefile`` in Windows Python 2.7
    """
    return os.path.normcase(os.path.normpath(os.path.realpath(path))) \
        == os.path.normcase(os.path.normpath(os.path.realpath(other)))


sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))
try:
    import zetup
except VersionConflict:
    egg_info = 'zetup.egg-info'
    dist = get_distribution('zetup')
    if samefile(
            dist.location, os.path.dirname(os.path.realpath(__file__))
    ) and os.path.exists(egg_info):
        print("zetup: Removing possibly outdated %s/" % egg_info,
              # don't pollute stdout
              file=sys.stderr)
        for fname in os.listdir(egg_info):
            os.remove(os.path.join(egg_info, fname))
        os.rmdir(egg_info)
        # when run via pip, the egg-info is still referenced by setuptools,
        # which would try to read the contents
        for keys in working_set.entry_keys.values():
            if 'zetup' in keys:
                keys.remove('zetup')
        del working_set.by_key['zetup']


from zetup import Zetup, DistributionNotFound, VersionConflict
try:
    from zetup.commands import make, pytest, tox, conda
except (ImportError, DistributionNotFound, VersionConflict):
    # ==> no zetup commands available
    # standard setup commands work anyway
    pass


# setup_req = 'setuptools >= 15.0'
# try:
#     get_distribution(setup_req)
# except VersionConflict:
#     for mod in ['setuptools', 'pkg_resources']:
#         for name, _ in list(sys.modules.items()):
#             if name == mod or name.startswith(mod + '.'):
#                 del sys.modules[name]
#     sys.path.insert(0, Distribution().fetch_build_egg(setup_req))


zfg = Zetup()
zetup.requires.Requirements('setuptools >= 36.2', zfg=zfg).check()

setup = zfg.setup
setup['package_data']['zetup.commands.make'] = [
    'templates/*.jinja',
    'templates/package/*.jinja',
]
setup()



#!python

# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

import sys
from itertools import chain
from argparse import ArgumentParser
import distutils.command

import zetup.commands
from zetup.commands import ZetupCommandError, \
    make, dev, del_, test, pytest, tox, conda
from zetup.process import call
from zetup.zetup import ZetupConfigNotFound


EXTERNAL_COMMANDS = []

COMMANDS = sorted(chain(
    distutils.command.__all__,
    zetup.commands.COMMANDS,
    zetup.Zetup.COMMANDS,
    EXTERNAL_COMMANDS,
))

PARSER = ArgumentParser()
PARSER.add_argument(
    'cmd', choices=COMMANDS,
    help="command",
)


def run(argv=None, cmd=None):
    """Run the **zetup** script.

    - If no `argv` is given, arguments are taken from ``sys.argv``.
    - Optionally takes an explicit zetup `cmd` not contained in `argv`.
    """
    argv = sys.argv[1:] if argv is None else list(argv)
    if cmd:
        argv.insert(1, str(cmd))
    args = PARSER.parse_args(argv)

    exit_status = 0 # exit status of this script
    try:
        zfg = zetup.Zetup()
    except ZetupConfigNotFound as no_zfg:
        try:
            cmdfunc = zetup.commands.COMMANDS[args.cmd]
        except KeyError:
            raise no_zfg
    else:
        if args.cmd in zfg.COMMANDS:
            cmdfunc = getattr(zfg, args.cmd)
        else: # ==> standard setup command
            sys.exit(zfg(subprocess=True))

    try:
        exit_status = cmdfunc()
    except ZetupCommandError as exc:
        print("Error: %s" % exc, file=sys.stderr)
        exit_status = 1
    else:
        try: # return value can be more than just a status number
            exit_status = exit_status.status
        except AttributeError:
            pass

    sys.exit(exit_status or 0)


def zake(argv=None):
    """Convenience runner for **zetup make** command.
    """
    run(argv, cmd='make')


def zev(argv=None):
    """Convenience runner for **zetup dev** command.
    """
    run(argv, cmd='dev')


def zel(argv=None):
    """Convenience runner for **zetup del** command.
    """
    run(argv, cmd='del')


def zest(argv=None):
    """Convenience runner for **zetup test** command.
    """
    run(argv, cmd='test')


def zox(argv=None):
    """Convenience runner for **zetup tox** command.
    """
    run(argv, cmd='tox')


if __name__ == '__main__':
    run()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.commands.delete

Defines ``zetup del`` command.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
import sys
import os

import pkg_resources
import pip

from path import Path

from zetup.zetup import Zetup
from zetup.commands.command import command
from zetup.conda import conda

__all__ = ['del_']


@Zetup.command(name='del')
@command(name='del')
def del_(zfg, args=None):
    """Delete project from python environment.
    """
    try:  # check for conda
        conda_info = conda.info()
    except OSError:  # ==> no conda
        pass
    else:
        # are we in a conda environment?
        if any(Path(conda_info[key]).samefile(sys.prefix)
               for key in ['root_prefix', 'default_prefix']
        # and is project installed via conda?
        ) and conda.list('--no-pip', '--full-name', zfg.NAME):
            # then also remove it via conda
            status = conda.remove(zfg.NAME, json=False)
            if status:  # ==> error
                return status
    # is there some project (develop) install (left) to be removed via pip?
    while True:
        try:
            # always use a refreshed working set
            # (interface to installed python package distributions)
            dist = pkg_resources.WorkingSet().by_key[zfg.NAME]
        except KeyError:  # ==> nothing left to uninstall
            break
        status = pip.main(['uninstall', zfg.NAME, '--yes'])
        if status:  # ==> error
            return status
        root = Path(dist.location)
        if root.exists() and root.samefile(zfg.ZETUP_DIR):
            # pip doesn't remove local .egg-info/ dirs of develop installs
            egg_info = Path(dist._provider.egg_info).realpath()
            print("zetup: Removing %s%s" % (egg_info, os.path.sep))
            egg_info.rmtree()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

__all__ = ['load_zetup_config']

import sys
import os
import re
from textwrap import dedent
from collections import OrderedDict
from warnings import warn
if sys.version_info[0] == 3:
    from configparser import ConfigParser
    # Just for simpler PY2/3 compatible code:
    unicode = str
else:
    from ConfigParser import ConfigParser

from .version import Version
from .requires import Requirements
from .extras import Extras
from .dist import Distribution
from .package import Packages
from .notebook import Notebook
from .resolve import resolve
from .error import ZetupError


TRUE = True, 'true', 'yes'
FALSE = False, 'false', 'no'


CONFIG_FILE_NAMES = ['zetuprc', 'zetup.cfg', 'zetup.ini']


class ZetupConfigNotFound(ZetupError):
    pass


def load_zetup_config(path, zfg):
    """Load zetup config from directory in `path`
       and store keywords as attributes to `zfg` object.
    """
    zfg.ZETUP_DIR = path

    # Read the zetup config...
    config = ConfigParser()
    for fname in CONFIG_FILE_NAMES:
        zfg.ZETUP_FILE = os.path.join(zfg.ZETUP_DIR, fname)
        if config.read(zfg.ZETUP_FILE):
            ##TODO: No print if run from installed package (under pkg/zetup/):
            ## print("zetup: Using config from %s" % fname)

            # The config file will be installed as pkg.zetup package_data:
            zfg.ZETUP_DATA = [fname]
            break
    else:
        raise ZetupConfigNotFound(
          "No zetup config file found in %s" % repr(path)
          + " (need %s)" % " or ".join([
            ", ".join(map(repr, CONFIG_FILE_NAMES[:-1])),
            repr(CONFIG_FILE_NAMES[-1])]))

    #... and store all setup options in UPPERCASE vars...
    zfg.NAME = config.sections()[0]

    # get a section dictionary with normalized option names as keys
    # and stripped value strings
    config = {re.sub(r'[^a-z0-9]', '', option.lower()): value.strip()
              for option, value in config.items(zfg.NAME)}

    zfg.TITLE = config.get('title', zfg.NAME)
    zfg.DESCRIPTION = config.get('description', '').replace('\n', ' ')

    zfg.AUTHOR = config.get('author')
    zfg.EMAIL = None
    if zfg.AUTHOR:
        match = re.match(r'^([^<]+)<([^>]+)>$', zfg.AUTHOR)
        if match:
            zfg.AUTHOR, zfg.EMAIL = map(str.strip, match.groups())
    zfg.URL = config.get('url')

    zfg.LICENSE = config.get('license')

    zfg.PYTHON = config.get('python', '').split()

    zfg.PACKAGES = config.get(
        'packages', Packages([], root=zfg.ZETUP_DIR, zfg=zfg))
    if zfg.PACKAGES:
        # First should be the root package
        zfg.PACKAGES = Packages(zfg.PACKAGES, root=zfg.ZETUP_DIR, zfg=zfg)
    elif os.path.isdir(os.path.join(zfg.ZETUP_DIR, *zfg.NAME.split('.'))):
        # Just assume distribution name == root package name
        zfg.PACKAGES = Packages([zfg.NAME], root=zfg.ZETUP_DIR, zfg=zfg)

    zfg.MODULES = (
        config.get('modules', '') or config.get('pymodules', '')
    ).split()

    zfg.ZETUP_CONFIG_PACKAGE = config.get('zetupconfigpackage')
    if zfg.ZETUP_CONFIG_PACKAGE:
        if zfg.ZETUP_CONFIG_PACKAGE in TRUE:
            zfg.ZETUP_CONFIG_PACKAGE = zfg.PACKAGES.main + '.zetup_config'
        elif zfg.ZETUP_CONFIG_PACKAGE in FALSE:
            zfg.ZETUP_CONFIG_PACKAGE = False
        # else it defines a custom package

    zfg.ZETUP_CONFIG_MODULE = config.get('zetupconfigmodule', 'yes')
    if zfg.ZETUP_CONFIG_MODULE:
        if zfg.ZETUP_CONFIG_MODULE in TRUE:
            if not zfg.PACKAGES:
                raise ZetupError(
                    "Can't add a default .zetup_config submodule"
                    " if no package is defined.")
            zfg.ZETUP_CONFIG_MODULE = zfg.PACKAGES.main + '.zetup_config'
        elif zfg.ZETUP_CONFIG_MODULE in FALSE:
            zfg.ZETUP_CONFIG_MODULE = False
        # else it defines a custom module

    zfg.SCRIPTS = config.get('scripts')
    if zfg.SCRIPTS:
        lines = list(map(str.strip, zfg.SCRIPTS.split('\n')))
        zfg.SCRIPTS = {}
        for line in lines:
            if not line:
                continue
            name, source = map(str.strip, line.split(':', 1))
            zfg.SCRIPTS[name] = source

    zfg.SETUP_KEYWORDS = config.get('setupkeywords')
    if zfg.SETUP_KEYWORDS:
        lines = list(map(str.strip, zfg.SETUP_KEYWORDS.split('\n')))
        zfg.SETUP_KEYWORDS = {}
        for line in lines:
            if not line:
                continue
            name, source = map(str.strip, line.split(':', 1))
            zfg.SETUP_KEYWORDS[name] = source

    zfg.ZETUP_CONFIG_HOOKS = config.get('zetupconfighooks', '').split()

    zfg.SETUP_HOOKS = config.get('setuphooks', '').split()

    zfg.NO_MAKE = config.get('nomake', '').split()

    zfg.KEEP_MADE = config.get('keepmade', '').split()

    zfg.FORCE_MAKE = config.get('forcemake', True)
    if zfg.FORCE_MAKE is not True:
        if zfg.FORCE_MAKE in TRUE:
            zfg.FORCE_MAKE = True
        elif zfg.FORCE_MAKE in FALSE:
            zfg.FORCE_MAKE = False
        else:
            raise ZetupConfigError(
                "Invalid value for 'force make' option: %s"
                % zfg.FORCE_MAKE)

    zfg.TEST_COMMANDS = list(filter(None, map(
        str.strip, config.get('testcommands', 'py.test -v test').split('\n')
    )))

    # get all non-empty classifier lines
    # (lines starting with :: are interpreted as continuation)
    zfg.CLASSIFIERS = list(filter(None, (
        line.strip() for line in re.sub(
            '\n\w*::', ' ::', config.get('classifiers', '').strip()
        ).split('\n'))))
    zfg.CLASSIFIERS.append('Programming Language :: Python')
    for pyversion in zfg.PYTHON:
        zfg.CLASSIFIERS.append(
            'Programming Language :: Python :: ' + pyversion)

    zfg.KEYWORDS = config.get('keywords', '').split()
    if any(pyversion.startswith('3') for pyversion in zfg.PYTHON):
        zfg.KEYWORDS.append('python3')

    # Parse VERSION and requirements files
    #  and add them to pkg.zetup package_data...
    zfg.ZETUP_DATA += ['VERSION', 'requirements.txt']

    zfg.VERSION_FILE = os.path.join(zfg.ZETUP_DIR, 'VERSION')
    if os.path.exists(zfg.VERSION_FILE):
        zfg.in_repo = False
        zfg.VERSION = open(zfg.VERSION_FILE).read().strip()
    else:
        zfg.in_repo = True
        zfg.VERSION_FILE = None
        try:
            import setuptools_scm
        except ImportError:
            warn(dedent(
                """No 'setuptools_scm' package found.
                   Zetup needs it to get project version from repository.
                """))
            zfg.VERSION = None
        else:
            version = setuptools_scm.get_version(root=zfg.ZETUP_DIR)
            # the hyphen-revision-hash part after .dev# version strings
            # results in wrong version comparisons
            # via pkg_resources.parse_version()
            zfg.VERSION = version and re.split('[-+]', version)[0]
    if zfg.VERSION:
        zfg.VERSION = Version(zfg.VERSION)

    zfg.DISTRIBUTION = Distribution(zfg)

    req_setup_txt = os.path.join(zfg.ZETUP_DIR, 'requirements.setup.txt')
    if os.path.exists(req_setup_txt):
        zfg.SETUP_REQUIRES = Requirements(
            open(req_setup_txt).read(), zfg=zfg)
    else:
        zfg.SETUP_REQUIRES = None

    req_txt = os.path.join(zfg.ZETUP_DIR, 'requirements.txt')
    if os.path.exists(req_txt):
        zfg.REQUIRES = Requirements(open(req_txt).read(), zfg=zfg)
    else:
        zfg.REQUIRES = None

    # Look for optional extra requirements to use with setup's extras_require=
    zfg.EXTRAS = Extras(zfg=zfg)
    for fname in sorted(os.listdir(zfg.ZETUP_DIR)):
        match = re.match(r'^requirements\.(?P<name>[^\.]+)\.txt$', fname)
        if match:
            name = match.group('name')
            if name == 'setup':
                # already handled in SETUP_REQUIRES
                continue

            zfg.ZETUP_DATA.append(fname)

            zfg.EXTRAS[name] = Requirements(
                open(os.path.join(zfg.ZETUP_DIR, fname)).read())

    # Are there IPython notebooks?
    zfg.NOTEBOOKS = OrderedDict()
    for fname in sorted(os.listdir(zfg.ZETUP_DIR)):
        name, ext = os.path.splitext(fname)
        if ext == '.ipynb':
            if name == 'README':
                zfg.ZETUP_DATA.append(fname)
            zfg.NOTEBOOKS[name] \
                = Notebook(os.path.join(zfg.ZETUP_DIR, fname))

    # finally resolve setup requirements...
    if zfg.SETUP_REQUIRES:
        resolve(zfg.SETUP_REQUIRES)

    # ... and run any custom zetup config hooks
    if zfg.ZETUP_CONFIG_HOOKS:
        sys.path.insert(0, zfg.ZETUP_DIR)
        for hook in zfg.ZETUP_CONFIG_HOOKS:
            modname, funcname = hook.split(':')
            mod = __import__(modname)
            for subname in modname.split('.')[1:]:
                mod = getattr(mod, subname)
            func = getattr(mod, funcname)
            func(zfg)
